"""Process the time delta between the last time played."""

# pylint: disable=duplicate-code,too-many-branches

import datetime
import functools

import pandas as pd
from tqdm import tqdm

from .columns import DELIMITER
from .identifier import Identifier


def lastplayed_process(
    df: pd.DataFrame, identifiers: list[Identifier], dt_column: str
) -> pd.DataFrame:
    """Process a dataframe for last played."""
    tqdm.pandas(desc="Last Played Features")
    last_identifier_dts: dict[str, datetime.datetime | None] = {}
    first_identifier_dts: dict[str, datetime.datetime] = {}

    def record_time(
        row: pd.Series,
        identifiers: list[Identifier],
        dt_column: str,
    ) -> pd.Series:
        nonlocal last_identifier_dts
        nonlocal first_identifier_dts

        dt = row[dt_column]
        for identifier in identifiers:
            if identifier.column not in row:
                continue
            identifier_id = row[identifier.column]
            if pd.isnull(identifier_id):
                continue
            if not isinstance(identifier_id, str):
                continue
            key = "_".join([str(identifier.entity_type), identifier_id])
            last_dt = last_identifier_dts.get(key)
            if last_dt is not None and dt is not None:
                row[DELIMITER.join([identifier.column_prefix, "lastplayeddays"])] = (
                    dt - last_dt
                ).days
            last_identifier_dts[key] = dt
            first_dt = first_identifier_dts.get(key)
            if first_dt is not None and dt is not None:
                row[DELIMITER.join([identifier.column_prefix, "firstplayeddays"])] = (
                    dt - first_dt
                ).days
            elif first_dt is None and dt is not None:
                first_identifier_dts[key] = dt
        return row

    return df.progress_apply(
        functools.partial(
            record_time,
            identifiers=identifiers,
            dt_column=dt_column,
        ),
        axis=1,
    ).copy()  # type: ignore



"""A process function for determing offensive efficiency of entities."""

import functools

import pandas as pd
from tqdm import tqdm

from .cache import MEMORY
from .columns import DELIMITER
from .identifier import Identifier

OFFENSIVE_EFFICIENCY_COLUMN = "offensiveefficiency"


@MEMORY.cache
def _record_offensive_efficiency(
    row: pd.Series, identifiers: list[Identifier]
) -> pd.Series:
    for identifier in identifiers:
        if identifier.field_goals_column is None:
            continue
        field_goals_value = row[identifier.field_goals_column]
        if pd.isnull(field_goals_value):
            continue
        field_goals = float(field_goals_value)
        if identifier.assists_column is None:
            continue
        assists_value = row[identifier.assists_column]
        if pd.isnull(assists_value):
            continue
        assists = float(assists_value)
        if identifier.field_goals_attempted_column is None:
            continue
        field_goals_attempted_value = row[identifier.field_goals_attempted_column]
        if pd.isnull(field_goals_attempted_value):
            continue
        field_goals_attempted = float(row[identifier.field_goals_attempted_column])
        if identifier.offensive_rebounds_column is None:
            continue
        offensive_rebounds_value = row[identifier.offensive_rebounds_column]
        if pd.isnull(offensive_rebounds_value):
            continue
        offensive_rebounds = float(offensive_rebounds_value)
        if identifier.turnovers_column is None:
            continue
        turnovers_value = row[identifier.turnovers_column]
        if pd.isnull(turnovers_value):
            continue
        turnovers = float(turnovers_value)
        offensive_efficiency_column = DELIMITER.join(
            [identifier.column_prefix, OFFENSIVE_EFFICIENCY_COLUMN]
        )
        row[offensive_efficiency_column] = (field_goals + assists) / (
            field_goals_attempted - offensive_rebounds + assists + turnovers
        )
        if offensive_efficiency_column not in identifier.feature_columns:
            identifier.feature_columns.append(offensive_efficiency_column)
    return row


def offensive_efficiency_process(
    df: pd.DataFrame, identifiers: list[Identifier]
) -> pd.DataFrame:
    """Process a dataframe for offensive efficiency."""
    tqdm.pandas(desc="Offensive Efficiency Features")

    return df.progress_apply(
        functools.partial(
            _record_offensive_efficiency,
            identifiers=identifiers,
        ),
        axis=1,
    ).copy()  # type: ignore



"""Calculate win features."""

# pylint: disable=too-many-locals,too-many-branches

from warnings import simplefilter

import pandas as pd
import tqdm

from .columns import DELIMITER
from .identifier import Identifier

WINS_COLUMN = "wins"


def win_process(df: pd.DataFrame, identifiers: list[Identifier]) -> pd.DataFrame:
    """Process wins between teams."""
    simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
    df_dict: dict[str, list[float | None]] = {}
    df_cols = df.columns.values.tolist()
    wins_dict: dict[str, float] = {}

    written_columns = set()
    for row in tqdm.tqdm(
        df.itertuples(name=None), desc="Win Processing", total=len(df)
    ):
        row_dict = {x: row[count + 1] for count, x in enumerate(df_cols)}

        # Write lagged data
        for identifier in identifiers:
            identifier_id = row_dict.get(identifier.column)
            if identifier_id is None:
                continue
            key = DELIMITER.join([identifier.entity_type, identifier_id])
            lagged_value = wins_dict.get(key)
            if lagged_value is None:
                continue
            col = DELIMITER.join([identifier.column_prefix, WINS_COLUMN])
            if col not in df_dict:
                df_dict[col] = [None for _ in range(len(df))]
            df_dict[col][row[0]] = lagged_value
            written_columns.add(col)
            del wins_dict[key]

        # Determine max points
        max_points = None
        for identifier in identifiers:
            points_col = identifier.points_column
            if points_col is None:
                continue
            points = row_dict[points_col]
            if points is None:
                continue
            if max_points is None:
                max_points = points
            else:
                max_points = max(max_points, points)

        # Write the next values
        if max_points is not None:
            for identifier in identifiers:
                identifier_id = row_dict.get(identifier.column)
                if identifier_id is None:
                    continue
                key = DELIMITER.join([identifier.entity_type, identifier_id])
                points_col = identifier.points_column
                if points_col is None:
                    continue
                points = row_dict[points_col]
                if points is None:
                    continue
                wins_dict[key] = float(points == max_points)

    for column in written_columns:
        df[column] = df_dict[column]

    # Add new feature columns
    for identifier in identifiers:
        col = DELIMITER.join([identifier.column_prefix, WINS_COLUMN])
        if col in written_columns:
            identifier.feature_columns.append(col)

    return df[sorted(df.columns.values.tolist())].copy()



"""Calculate margin features."""

# pylint: disable=too-many-locals,too-many-branches,duplicate-code

import sys
from warnings import simplefilter

import pandas as pd
import tqdm

from .columns import DELIMITER
from .identifier import Identifier


def margin_process(df: pd.DataFrame, identifiers: list[Identifier]) -> pd.DataFrame:
    """Process margins between teams."""
    simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
    df_dict: dict[str, list[float | None]] = {}
    df_cols = df.columns.values.tolist()
    identifiers_ts: dict[str, dict[str, float]] = {}

    written_columns = set()
    for row in tqdm.tqdm(
        df.itertuples(name=None), desc="Margin Processing", total=len(df)
    ):
        row_dict = {x: row[count + 1] for count, x in enumerate(df_cols)}

        # Write lagged data
        for identifier in identifiers:
            identifier_id = row_dict.get(identifier.column)
            if identifier_id is None:
                continue
            key = DELIMITER.join([identifier.entity_type, identifier_id])
            lagged_values = identifiers_ts.get(key)
            if lagged_values is None:
                continue
            for k, v in lagged_values.items():
                col = identifier.column_prefix + k
                if col not in df_dict:
                    df_dict[col] = [None for _ in range(len(df))]
                df_dict[col][row[0]] = v
                written_columns.add(col)
            del identifiers_ts[key]

        # Find maximum value
        entity_dicts: dict[str, dict[str, float]] = {}
        for identifier in identifiers:
            identifier_id = row_dict.get(identifier.column)
            if identifier_id is None:
                continue
            entity_dict = entity_dicts.get(identifier.entity_type, {})
            for column in identifier.numeric_action_columns:
                value = row_dict.get(column)
                if value is None:
                    continue
                feature_column = column[len(identifier.column_prefix) :]
                entity_dict[feature_column] = max(
                    entity_dict.get(feature_column, sys.float_info.min), value
                )
            entity_dicts[identifier.entity_type] = entity_dict

        # Cache lagged data
        for identifier in identifiers:
            identifier_id = row_dict.get(identifier.column)
            if identifier_id is None:
                continue
            feature_dict = entity_dicts[identifier.entity_type]
            key = DELIMITER.join([identifier.entity_type, identifier_id])
            identifier_dict = {}
            for k, max_value in feature_dict.items():
                full_col = identifier.column_prefix + k
                if full_col not in row_dict:
                    continue
                value = row_dict[full_col]
                abs_col = DELIMITER.join([k, "margin", "absolute"])
                rel_col = DELIMITER.join([k, "margin", "relative"])
                identifier_dict[abs_col] = value - max_value
                identifier_dict[rel_col] = value / max_value
            identifiers_ts[key] = identifier_dict

    for column in written_columns:
        df[column] = df_dict[column]

    return df[sorted(df.columns.values.tolist())].copy()



"""Process a dataframe for its news information."""

# pylint: disable=too-many-locals,too-many-branches
import statistics
from warnings import simplefilter

import pandas as pd
import tqdm
from textfeats.process import process  # type: ignore

from .columns import DELIMITER
from .identifier import Identifier

COUNT_COLUMN = "count"
MENTIONS_COLUMN = "mentions"
EMBEDDING_COLUMN = "embedding"


def news_process(df: pd.DataFrame, identifiers: list[Identifier]) -> pd.DataFrame:
    """Process news features."""
    simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
    df_dict: dict[str, list[float | None]] = {}
    df_cols = df.columns.values.tolist()

    written_columns = set()
    for row in tqdm.tqdm(
        df.itertuples(name=None), desc="News Processing", total=len(df)
    ):
        row_dict = {x: row[count + 1] for count, x in enumerate(df_cols)}

        for identifier in identifiers:
            summary_columns = []
            for news in identifier.news:
                if news.summary_column not in row_dict:
                    continue
                if row_dict[news.summary_column] is None:
                    continue
                summary_columns.append(news.summary_column)

            # Calculate counts
            count_col = DELIMITER.join([identifier.column_prefix, "news", COUNT_COLUMN])
            if count_col not in df_dict:
                df_dict[count_col] = [None for _ in range(len(df))]
            df_dict[count_col][row[0]] = float(len(summary_columns))
            written_columns.add(count_col)

            # Calculate injury mentions
            summaries = [row_dict[x] for x in summary_columns]
            injury_mentions = 0
            for summary in summaries:
                if "injur" in summary:
                    injury_mentions += 1
            injury_mentions_column = DELIMITER.join(
                [identifier.column_prefix, "news", MENTIONS_COLUMN, "injury"]
            )
            if injury_mentions_column not in df_dict:
                df_dict[injury_mentions_column] = [None for _ in range(len(df))]
            df_dict[injury_mentions_column][row[0]] = float(injury_mentions)
            written_columns.add(injury_mentions_column)

            # Calculate embeddings
            news_df = process(
                pd.DataFrame(
                    data={k: v for k, v in row_dict.items() if k in summary_columns}
                )
            )
            news_df = news_df.drop(columns=summary_columns, errors="ignore")
            embedding_indexes = set()
            for column in news_df.columns.values.tolist():
                column_split = column.split("_")
                embedding_indexes.add(column_split[-1])
            for embedding_index in embedding_indexes:
                values = []
                for column in news_df.columns.values.tolist():
                    column_split = column.split("_")
                    if column_split[-1] == embedding_index:
                        values.append(news_df[column].tolist()[0])
                embedding_column = DELIMITER.join(
                    [
                        identifier.column_prefix,
                        "news",
                        EMBEDDING_COLUMN,
                        embedding_index,
                    ]
                )
                if embedding_column not in df_dict:
                    df_dict[embedding_column] = [None for _ in range(len(df))]
                df_dict[embedding_column][row[0]] = statistics.mean(values)
                written_columns.add(embedding_column)

    for column in written_columns:
        df[column] = df_dict[column]

    return df.copy()



"""Process a dataframe for its datetime information."""

from warnings import simplefilter

import pandas as pd
from feature_engine.datetime import DatetimeFeatures


def datetime_process(df: pd.DataFrame, dt_column: str) -> pd.DataFrame:
    """Process datetime features."""
    simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
    dtf = DatetimeFeatures(
        variables=[dt_column],
        features_to_extract="all",
        missing_values="ignore",
        drop_original=False,
        utc=True,
    )
    return dtf.fit_transform(df).copy()



"""Process the distance between two locations."""

# pylint: disable=duplicate-code,too-many-branches,too-many-locals,too-many-statements

import geopy.distance  # type: ignore
import pandas as pd
from tqdm import tqdm

from .columns import DELIMITER
from .entity_type import EntityType
from .identifier import Identifier


def distance_process(df: pd.DataFrame, identifiers: list[Identifier]) -> pd.DataFrame:
    """Process a dataframe for offensive efficiency."""
    last_identifier_locations: dict[str, tuple[float, float]] = {}
    team_identifiers = [x for x in identifiers if x.entity_type == EntityType.TEAM]
    player_identifiers = [x for x in identifiers if x.entity_type == EntityType.PLAYER]
    venue_identifiers = [x for x in identifiers if x.entity_type == EntityType.VENUE]
    df_dict: dict[str, list[float | None]] = {}
    df_cols = df.columns.values.tolist()

    written_columns = set()
    for row in tqdm(
        df.itertuples(name=None), desc="Distance Processing", total=len(df)
    ):
        row_dict = {x: row[count + 1] for count, x in enumerate(df_cols)}

        current_location = None
        for venue_identifier in venue_identifiers:
            if venue_identifier.latitude_column is None:
                continue
            if venue_identifier.latitude_column not in row_dict:
                continue
            latitude = row_dict[venue_identifier.latitude_column]
            if latitude is None:
                continue
            if venue_identifier.longitude_column is None:
                continue
            if venue_identifier.longitude_column not in row_dict:
                continue
            longitude = row_dict[venue_identifier.longitude_column]
            if longitude is None:
                continue
            current_location = (latitude, longitude)
        if current_location is None:
            continue

        for identifier in team_identifiers + player_identifiers:
            if identifier.column not in row_dict:
                continue
            identifier_id = row_dict[identifier.column]
            if identifier_id is None:
                continue
            if not isinstance(identifier_id, str):
                continue
            key = "_".join([str(identifier.entity_type), identifier_id])
            last_location = last_identifier_locations.get(key)
            if last_location is not None:
                latitude_diff_column = DELIMITER.join(
                    [identifier.column_prefix, "latitudediff"]
                )
                if latitude_diff_column not in df_dict:
                    df_dict[latitude_diff_column] = [None for _ in range(len(df))]
                written_columns.add(latitude_diff_column)
                df_dict[latitude_diff_column][row[0]] = abs(
                    last_location[0] - current_location[0]
                )
                longitude_diff_column = DELIMITER.join(
                    [identifier.column_prefix, "longitudediff"]
                )
                if longitude_diff_column not in df_dict:
                    df_dict[longitude_diff_column] = [None for _ in range(len(df))]
                written_columns.add(longitude_diff_column)
                df_dict[longitude_diff_column][row[0]] = abs(
                    last_location[1] - current_location[1]
                )
                distance_column = DELIMITER.join([identifier.column_prefix, "distance"])
                if distance_column not in df_dict:
                    df_dict[distance_column] = [None for _ in range(len(df))]
                df_dict[distance_column][row[0]] = geopy.distance.geodesic(
                    last_location, current_location
                ).km
                written_columns.add(distance_column)
            last_identifier_locations[key] = current_location

    for column in written_columns:
        df[column] = df_dict[column]

    return df[sorted(df.columns.values.tolist())].copy()



"""Processing for time series features."""

# pylint: disable=duplicate-code,too-many-branches,too-many-nested-blocks

import datetime
import multiprocessing
from warnings import simplefilter

import pandas as pd
from timeseriesfeatures.feature import FEATURE_TYPE_LAG  # type: ignore
from timeseriesfeatures.feature import (FEATURE_TYPE_ROLLING, VALUE_TYPE_DAYS,
                                        VALUE_TYPE_NONE, Feature)
from timeseriesfeatures.process import process  # type: ignore
from timeseriesfeatures.transform import Transform  # type: ignore
from tqdm import tqdm

from .columns import DELIMITER
from .entity_type import EntityType
from .identifier import Identifier

_COLUMN_PREFIX_COLUMN = "_column_prefix"


def _pool_process(
    identifier_id: str,
    df: pd.DataFrame,
    features: list[Feature],
    dt_column: str,
) -> tuple[str, pd.DataFrame, pd.Series]:
    original_identifier_df = df.copy()
    drop_columns = df.columns.values.tolist()
    if "" in drop_columns:
        drop_columns.remove("")
    drop_columns.remove(_COLUMN_PREFIX_COLUMN)
    return (
        identifier_id,
        process(df, features=features, on=dt_column).drop(columns=drop_columns).copy(),
        original_identifier_df[_COLUMN_PREFIX_COLUMN],
    )


def _extract_identifier_timeseries(
    df: pd.DataFrame, identifiers: list[Identifier], dt_column: str
) -> dict[str, pd.DataFrame]:
    tqdm.pandas(desc="Timeseries Progress")
    identifier_ts: dict[str, pd.DataFrame] = {}
    team_identifiers = [x for x in identifiers if x.entity_type == EntityType.TEAM]
    player_identifiers = [x for x in identifiers if x.entity_type == EntityType.PLAYER]
    relevant_identifiers = team_identifiers + player_identifiers

    def record_timeseries_features(row: pd.Series) -> pd.Series:
        nonlocal identifier_ts
        nonlocal relevant_identifiers

        for identifier in relevant_identifiers:
            if identifier.column not in row:
                continue
            identifier_id = row[identifier.column]
            if pd.isnull(identifier_id):
                continue
            key = DELIMITER.join([identifier.entity_type, identifier_id])
            identifier_df = identifier_ts.get(key, pd.DataFrame())
            identifier_df.loc[row.name, _COLUMN_PREFIX_COLUMN] = (  # type: ignore
                identifier.column_prefix
            )
            identifier_df.loc[row.name, dt_column] = row[dt_column]  # type: ignore
            for feature_column in identifier.numeric_action_columns:
                if feature_column not in row:
                    continue
                value = row[feature_column]
                if pd.isnull(value):
                    continue
                column = feature_column[len(identifier.column_prefix) :]
                if not column:
                    continue
                if column not in identifier_df:
                    identifier_df[column] = None
                identifier_df.loc[row.name, column] = value  # type: ignore
            identifier_ts[key] = identifier_df.infer_objects()

        return row

    df.progress_apply(record_timeseries_features, axis=1)  # type: ignore
    return identifier_ts


def _process_identifier_ts(
    identifier_ts: dict[str, pd.DataFrame],
    windows: list[datetime.timedelta | None],
    dt_column: str,
) -> dict[str, pd.DataFrame]:
    # pylint: disable=too-many-locals
    features = [
        Feature(
            feature_type=FEATURE_TYPE_LAG,
            columns=[],
            value1=1,
            transform=str(Transform.NONE),
        ),
        Feature(
            feature_type=FEATURE_TYPE_LAG,
            columns=[],
            value1=2,
            transform=str(Transform.NONE),
        ),
        Feature(
            feature_type=FEATURE_TYPE_LAG,
            columns=[],
            value1=4,
            transform=str(Transform.NONE),
        ),
        Feature(
            feature_type=FEATURE_TYPE_LAG,
            columns=[],
            value1=8,
            transform=str(Transform.NONE),
        ),
    ] + [
        Feature(
            feature_type=FEATURE_TYPE_ROLLING,
            columns=[],
            value1=VALUE_TYPE_NONE if x is None else VALUE_TYPE_DAYS,
            value2=None if x is None else x.days,
            transform=str(Transform.NONE),
        )
        for x in windows
    ]
    with multiprocessing.Pool() as pool:
        for identifier_id, identifier_df, column_prefix_series in pool.starmap(
            _pool_process,
            tqdm(
                [(k, v, features, dt_column) for k, v in identifier_ts.items()],
                desc="Timeseries Processing",
            ),
        ):
            identifier_ts[identifier_id] = identifier_df
            identifier_ts[identifier_id][_COLUMN_PREFIX_COLUMN] = column_prefix_series

    return identifier_ts


def _write_ts_features(
    df: pd.DataFrame,
    identifier_ts: dict[str, pd.DataFrame],
    dt_column: str,
) -> pd.DataFrame:
    df_dict = df.to_dict(orient="list")

    written_columns = set()
    for identifier_df in tqdm(
        identifier_ts.values(), desc="Writing Timeseries Features"
    ):
        for row in identifier_df.itertuples(name=None):
            row_dict = {
                x: row[count + 1]
                for count, x in enumerate(identifier_df.columns.values.tolist())
            }
            column_prefix = row_dict[_COLUMN_PREFIX_COLUMN]
            for column, value in row_dict.items():
                if column in {_COLUMN_PREFIX_COLUMN, dt_column, ""}:
                    continue
                key = column_prefix + column
                if key not in df_dict:
                    df_dict[key] = [None for _ in range(len(df))]
                df_dict[key][row[0]] = value
                written_columns.add(key)
    for column in written_columns:
        df[column] = df_dict[column]

    return df


def timeseries_process(
    df: pd.DataFrame,
    identifiers: list[Identifier],
    windows: list[datetime.timedelta | None],
    dt_column: str,
) -> pd.DataFrame:
    """Process a dataframe for its timeseries features."""
    # pylint: disable=too-many-locals,consider-using-dict-items,too-many-statements,duplicate-code
    tqdm.pandas(desc="Progress")
    simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

    # Write the columns to the dataframe ahead of time.
    df[_COLUMN_PREFIX_COLUMN] = None

    identifier_ts: dict[str, pd.DataFrame] = _extract_identifier_timeseries(
        df, identifiers, dt_column
    )
    identifier_ts = _process_identifier_ts(identifier_ts, windows, dt_column)
    df = _write_ts_features(df, identifier_ts, dt_column)
    return df.drop(columns=[_COLUMN_PREFIX_COLUMN]).copy()



"""A description of a news article in the dataframe."""

# pylint: disable=too-many-arguments,too-many-positional-arguments


class News:
    """A way to define a news article."""

    # pylint: disable=too-few-public-methods

    def __init__(
        self,
        title_column: str,
        published_column: str,
        summary_column: str,
        source_column: str,
    ):
        self.title_column = title_column
        self.published_column = published_column
        self.summary_column = summary_column
        self.source_column = source_column



"""The process for removing lookahead features."""

import pandas as pd

from .identifier import Identifier


def remove_process(df: pd.DataFrame, identifiers: list[Identifier]) -> pd.DataFrame:
    """Remove the features from the dataframe."""
    drop_columns: set[str] = set()
    for identifier in identifiers:
        for feature_col in identifier.feature_columns:
            drop_columns.add(feature_col)
        for bet in identifier.bets:
            drop_columns.add(bet.odds_column)
            drop_columns.add(bet.bookie_id_column)
            if bet.dt_column is not None:
                drop_columns.add(bet.dt_column)
            drop_columns.add(bet.bookie_name_column)
            drop_columns.add(bet.canonical_column)
        for news in identifier.news:
            drop_columns.add(news.title_column)
            drop_columns.add(news.published_column)
            drop_columns.add(news.summary_column)
            drop_columns.add(news.source_column)
    return df.drop(columns=list(drop_columns), errors="ignore").copy()



"""Process a dataframe for its ordinal information."""

import pandas as pd
from feature_engine.encoding import OrdinalEncoder


def ordinal_process(df: pd.DataFrame, categorical_features: set[str]) -> pd.DataFrame:
    """Process ordinal features."""
    if not categorical_features:
        return df
    true_categorical_features: list[str] = []
    for categorical_feature in categorical_features:
        if categorical_feature not in df.columns.values:
            continue
        true_categorical_features.append(categorical_feature)
        df[categorical_feature] = df[categorical_feature].astype("category")
    od = OrdinalEncoder(
        missing_values="ignore",
        encoding_method="arbitrary",
        variables=true_categorical_features,  # type: ignore
    )
    df = od.fit_transform(df)
    for categorical_feature in true_categorical_features:
        if categorical_feature not in df.columns.values:
            continue
        try:
            df[categorical_feature] = (
                df[categorical_feature].fillna(0).astype(int).astype("category")
            )
        except TypeError:
            pass
    return df.copy()



"""Tests for the bet process function."""
import datetime
import unittest
import os

import pandas as pd
from pandas.testing import assert_frame_equal

from sportsfeatures.bets_process import bet_process
from sportsfeatures.identifier import Identifier
from sportsfeatures.entity_type import EntityType
from sportsfeatures.bet import Bet


class TestBetProcess(unittest.TestCase):

    def setUp(self):
        self.dir = os.path.dirname(__file__)

    def test_bet_process(self):
        df = pd.read_csv(os.path.join(self.dir, "bets.csv"))
        identifiers = [
            Identifier(
                EntityType.TEAM,
                "teams/0/identifier",
                [],
                "teams/0",
                points_column="teams/0/points",
                bets=[Bet(
                    odds_column=f"teams/0/odds/{x}/odds",
                    bookie_id_column=f"teams/0/odds/{x}/bookie/name",
                    dt_column=f"teams/0/odds/{x}/dt",
                    canonical_column=f"teams/0/odds/{x}/canonical",
                    bookie_name_column=f"teams/0/odds/{x}/bookie/realname",
                ) for x in range(28)],
            ),
            Identifier(
                EntityType.TEAM,
                "teams/1/identifier",
                [],
                "teams/1",
                points_column="teams/1/points",
                bets=[Bet(
                    odds_column=f"teams/1/odds/{x}/odds",
                    bookie_id_column=f"teams/1/odds/{x}/bookie/name",
                    dt_column=f"teams/1/odds/{x}/dt",
                    canonical_column=f"teams/1/odds/{x}/canonical",
                    bookie_name_column=f"teams/1/odds/{x}/bookie/realname",
                ) for x in range(28)],
            ),
        ]
        new_df = bet_process(df, identifiers, "dt", True)
        odds = new_df["teams/0_odds"].to_list()
        self.assertListEqual(odds, [
            4.0,
            8.0,
            1.3333333333333333,
            2.203333333333333,
            1.3050000000000002,
            1.4849999999999999,
            2.675,
            1.47,
            2.25,
            1.03,
            1.51,
            2.05,
            2.65,
            3.35,
            2.1500000000000004,
            1.705,
            2.05,
            1.5150000000000001,
            1.49,
            5.5,
        ])
        odds = new_df["teams/1_odds"].to_list()
        self.assertListEqual(odds, [
            2.29,
            1.03,
            3.266666666666667,
            1.6266666666666667,
            3.2,
            2.425,
            1.41,
            2.4749999999999996,
            1.5550000000000002,
            8.25,
            2.375,
            1.68,
            1.435,
            1.28,
            1.62,
            2.0,
            1.68,
            2.375,
            2.425,
            1.105,
        ])



"""Caching functionality for sportsfeatures."""

# pylint: disable=consider-using-enumerate
import hashlib
import os

import pandas as pd
import pytest_is_running
from joblib import Memory  # type: ignore

from . import __VERSION__

_SPORTS_FEATURES_CACHE_FOLDER = ".sportsfeatures_" + __VERSION__
SPORTS_FEATURES_DF_FILENAME = "df.parquet.gzip"
MEMORY = (
    Memory(".sportsfeatures_joblib_cache", verbose=0)
    if not pytest_is_running
    else Memory()
)


def sportsfeatures_cache_folder() -> str:
    """Return a valid cache folder."""
    if not os.path.exists(_SPORTS_FEATURES_CACHE_FOLDER):
        try:
            os.mkdir(_SPORTS_FEATURES_CACHE_FOLDER)
        except FileExistsError:
            pass
    return _SPORTS_FEATURES_CACHE_FOLDER


def _first_difference_position(df1: pd.DataFrame, df2: pd.DataFrame) -> int | None:
    # Align by index and columns to avoid misalignment issues
    df1, df2 = df1.align(df2)
    # Compare element-wise
    try:
        diff = (df1 != df2).values
        # Iterate by row position
        for i in range(len(diff)):
            if diff[i].any():
                if i == 0:
                    return None
                return i
        return len(diff)  # No difference found
    except TypeError:
        return None


def find_best_cache(cache_name: str, df: pd.DataFrame) -> tuple[str | None, int]:
    """Finds the best cached fit."""
    cache_folder = os.path.join(sportsfeatures_cache_folder(), cache_name)
    if not os.path.exists(cache_folder):
        os.mkdir(cache_folder)
    max_idx = -1
    cache_path = None
    for item in os.listdir(cache_folder):
        full_path = os.path.join(cache_folder, item)
        if not os.path.isdir(full_path):
            continue
        pd_file = os.path.join(full_path, SPORTS_FEATURES_DF_FILENAME)
        cached_df = pd.read_parquet(pd_file)
        diff_idx = _first_difference_position(df, cached_df)
        if diff_idx is not None:
            if diff_idx > max_idx:
                max_idx = diff_idx
                cache_path = full_path
    return cache_path, max_idx


def create_cache(cache_name: str, df: pd.DataFrame) -> str:
    """Creates a cache folder and stores the df."""
    df_hash = hashlib.sha256(df.to_csv().encode()).hexdigest()
    cache_folder = os.path.join(sportsfeatures_cache_folder(), cache_name, df_hash)
    os.makedirs(cache_folder, exist_ok=True)
    df.to_parquet(os.path.join(cache_folder, SPORTS_FEATURES_DF_FILENAME))
    return cache_folder



"""Process the current dataframe by adding skill features."""

# pylint: disable=duplicate-code,too-many-locals,too-many-statements

import datetime
import functools
import os

import numpy as np
import pandas as pd
from tqdm import tqdm

from .cache import MEMORY, create_cache, find_best_cache
from .columns import DELIMITER
from .entity_type import EntityType
from .identifier import Identifier
from .windowed_rating import WindowedRating

SKILL_COLUMN_PREFIX = "skill"
SKILL_MU_COLUMN = "mu"
SKILL_SIGMA_COLUMN = "sigma"
SKILL_RANKING_COLUMN = "ranking"
SKILL_PROBABILITY_COLUMN = "probability"
TIME_SLICE_ALL = "all"
_SKILL_CACHE_NAME = "skill"
_RESULTS_PARQUET_FILENAME = "results.parquet"


@MEMORY.cache
def _calculate_skills(
    row: pd.Series,
    rating_windows: list[WindowedRating],
    team_identifiers: list[Identifier],
    player_identifiers: list[Identifier],
) -> pd.Series:
    for rating_window in rating_windows:
        window_id = (
            TIME_SLICE_ALL
            if rating_window.window is None
            else f"window{rating_window.window.days}"
        )
        team_result, player_result = rating_window.add(
            row, team_identifiers, player_identifiers
        )
        for team_identifier in team_identifiers:
            if team_identifier.column not in row:
                continue
            team_id = row[team_identifier.column]
            if pd.isnull(team_id):
                continue
            if team_id in team_result:
                rating, ranking, prob = team_result[team_id]
                window_prefix = DELIMITER.join(
                    [team_identifier.column_prefix, SKILL_COLUMN_PREFIX, window_id]
                )
                row[DELIMITER.join([window_prefix, SKILL_MU_COLUMN])] = rating.mu
                row[DELIMITER.join([window_prefix, SKILL_SIGMA_COLUMN])] = rating.sigma
                row[DELIMITER.join([window_prefix, SKILL_RANKING_COLUMN])] = ranking
                row[DELIMITER.join([window_prefix, SKILL_PROBABILITY_COLUMN])] = prob
            for player_identifier in player_identifiers:
                if player_identifier.column not in row:
                    continue
                player_id = row[player_identifier.column]
                if pd.isnull(player_id):
                    continue
                if player_id in player_result:
                    rating, ranking, prob = team_result[team_id]
                    window_prefix = DELIMITER.join(
                        [
                            player_identifier.column_prefix,
                            SKILL_COLUMN_PREFIX,
                            window_id,
                        ]
                    )
                    row[DELIMITER.join([window_prefix, SKILL_MU_COLUMN])] = rating.mu
                    row[DELIMITER.join([window_prefix, SKILL_SIGMA_COLUMN])] = (
                        rating.sigma
                    )
                    row[DELIMITER.join([window_prefix, SKILL_RANKING_COLUMN])] = ranking
                    row[DELIMITER.join([window_prefix, SKILL_PROBABILITY_COLUMN])] = (
                        prob
                    )

    return row


def skill_process(
    df: pd.DataFrame,
    dt_column: str,
    identifiers: list[Identifier],
    windows: list[datetime.timedelta | None],
) -> pd.DataFrame:
    """Add skill features to the dataframe."""
    tqdm.pandas(desc="Skill Features")
    original_df = df.copy()

    team_identifiers = [x for x in identifiers if x.entity_type == EntityType.TEAM]
    player_identifiers = [x for x in identifiers if x.entity_type == EntityType.PLAYER]
    rating_windows = [WindowedRating(x, dt_column) for x in windows]

    cache_folder, idx = find_best_cache(_SKILL_CACHE_NAME, df)
    if cache_folder is not None:
        load_success = False
        for rating_window in rating_windows:
            load_success = rating_window.load(cache_folder)
            if not load_success:
                break
        if not load_success:
            for rating_window in rating_windows:
                rating_window.reset()
            cache_folder = None
        else:
            df_cache = pd.read_parquet(
                os.path.join(cache_folder, _RESULTS_PARQUET_FILENAME)
            )
            if len(df_cache) == len(df):
                return df
            df = df.join(
                df_cache.drop(
                    columns=[col for col in df_cache.columns if col in df.columns]
                ),
                how="left",
            )

    if cache_folder is not None:
        processed_ilocs = np.arange(idx)
        all_ilocs = np.arange(len(df))
        unprocessed_ilocs = np.setdiff1d(all_ilocs, processed_ilocs)
        df.iloc[unprocessed_ilocs] = df.iloc[unprocessed_ilocs].progress_apply(
            functools.partial(
                _calculate_skills,
                rating_windows=rating_windows,
                team_identifiers=team_identifiers,
                player_identifiers=player_identifiers,
            ),
            axis=1,
        )  # type: ignore
    else:
        df = df.progress_apply(
            functools.partial(
                _calculate_skills,
                rating_windows=rating_windows,
                team_identifiers=team_identifiers,
                player_identifiers=player_identifiers,
            ),
            axis=1,
        )  # type: ignore

    cache_folder = create_cache(_SKILL_CACHE_NAME, original_df)
    for rating_window in rating_windows:
        rating_window.save(cache_folder)
    df.to_parquet(os.path.join(cache_folder, _RESULTS_PARQUET_FILENAME))

    return df.copy()
